/**
 * Author......: Jens Steube <jens.steube@gmail.com>
 * License.....: MIT
 */

#define _MYSQL323_

#define NEW_SIMD_CODE

#include "include/constants.h"
#include "include/kernel_vendor.h"

#define DGST_R0 0
#define DGST_R1 1
#define DGST_R2 2
#define DGST_R3 3

#include "include/kernel_functions.c"
#include "OpenCL/types_ocl.c"
#include "OpenCL/common.c"
#include "OpenCL/simd.c"

__kernel void m00200_m04 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
  /**
   * modifier
   */

  const u32 lid = get_local_id (0);

  /**
   * base
   */

  const u32 gid = get_global_id (0);

  if (gid >= gid_max) return;

  u32 pws0[4] = { 0 };
  u32 pws1[4] = { 0 };

  pws0[0] = pws[gid].i[0];
  pws0[1] = pws[gid].i[1];
  pws0[2] = pws[gid].i[2];
  pws0[3] = pws[gid].i[3];
  pws1[0] = pws[gid].i[4];
  pws1[1] = pws[gid].i[5];
  pws1[2] = pws[gid].i[6];
  pws1[3] = pws[gid].i[7];

  const u32 pw_l_len = pws[gid].pw_len;

  /**
   * loop
   */

  for (u32 il_pos = 0; il_pos < combs_cnt; il_pos += VECT_SIZE)
  {
    const u32x pw_r_len = pwlenx_create_combt (combs_buf, il_pos);

    const u32x pw_len = pw_l_len + pw_r_len;

    u32x wordr0[4] = { 0 };
    u32x wordr1[4] = { 0 };
    u32x wordr2[4] = { 0 };
    u32x wordr3[4] = { 0 };

    wordr0[0] = ix_create_combt (combs_buf, il_pos, 0);
    wordr0[1] = ix_create_combt (combs_buf, il_pos, 1);
    wordr0[2] = ix_create_combt (combs_buf, il_pos, 2);
    wordr0[3] = ix_create_combt (combs_buf, il_pos, 3);
    wordr1[0] = ix_create_combt (combs_buf, il_pos, 4);
    wordr1[1] = ix_create_combt (combs_buf, il_pos, 5);
    wordr1[2] = ix_create_combt (combs_buf, il_pos, 6);
    wordr1[3] = ix_create_combt (combs_buf, il_pos, 7);

    if (combs_mode == COMBINATOR_MODE_BASE_LEFT)
    {
      switch_buffer_by_offset_le (wordr0, wordr1, wordr2, wordr3, pw_l_len);
    }

    u32x w_t[16];

    w_t[ 0] = wordl0[0] | wordr0[0];
    w_t[ 1] = wordl0[1] | wordr0[1];
    w_t[ 2] = wordl0[2] | wordr0[2];
    w_t[ 3] = wordl0[3] | wordr0[3];
    w_t[ 4] = wordl1[0] | wordr1[0];
    w_t[ 5] = wordl1[1] | wordr1[1];
    w_t[ 6] = wordl1[2] | wordr1[2];
    w_t[ 7] = wordl1[3] | wordr1[3];
    w_t[ 8] = wordl2[0] | wordr2[0];
    w_t[ 9] = wordl2[1] | wordr2[1];
    w_t[10] = wordl2[2] | wordr2[2];
    w_t[11] = wordl2[3] | wordr2[3];
    w_t[12] = wordl3[0] | wordr3[0];
    w_t[13] = wordl3[1] | wordr3[1];
    w_t[14] = wordl3[2] | wordr3[2];
    w_t[15] = 0;

    u32x a = MYSQL323_A;
    u32x b = MYSQL323_B;
    u32x c = 0;
    u32x d = 0;

    u32 add = 7;

    #define ROUND(v)                              \
    {                                             \
      a ^= (((a & 0x3f) + add) * (v)) + (a << 8); \
      b += (b << 8) ^ a;                          \
      add += v;                                   \
    }

    int i;
    int j;

    for (i = 0, j = 0; i <= (int) pw_len - 4; i += 4, j += 1)
    {
      const u32 wj = w_t[j];

      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
      ROUND ((wj >> 16) & 0xff);
      ROUND ((wj >> 24) & 0xff);
    }

    const u32 wj = w_t[j];

    const u32 left = pw_len - i;

    if (left == 3)
    {
      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
      ROUND ((wj >> 16) & 0xff);
    }
    else if (left == 2)
    {
      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
    }
    else if (left == 1)
    {
      ROUND ((wj >>  0) & 0xff);
    }

    a &= 0x7fffffff;
    b &= 0x7fffffff;

    COMPARE_M_SIMD (a, b, c, d);
  }
}

__kernel void m00200_m08 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
}

__kernel void m00200_m16 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
}

__kernel void m00200_s04 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
  /**
   * modifier
   */

  const u32 lid = get_local_id (0);

  /**
   * base
   */

  const u32 gid = get_global_id (0);

  if (gid >= gid_max) return;

  u32 pws0[4] = { 0 };
  u32 pws1[4] = { 0 };

  pws0[0] = pws[gid].i[0];
  pws0[1] = pws[gid].i[1];
  pws0[2] = pws[gid].i[2];
  pws0[3] = pws[gid].i[3];
  pws1[0] = pws[gid].i[4];
  pws1[1] = pws[gid].i[5];
  pws1[2] = pws[gid].i[6];
  pws1[3] = pws[gid].i[7];

  const u32 pw_l_len = pws[gid].pw_len;

  /**
   * digest
   */

  const u32 search[4] =
  {
    digests_buf[digests_offset].digest_buf[DGST_R0],
    digests_buf[digests_offset].digest_buf[DGST_R1],
    digests_buf[digests_offset].digest_buf[DGST_R2],
    digests_buf[digests_offset].digest_buf[DGST_R3]
  };

  /**
   * loop
   */

  for (u32 il_pos = 0; il_pos < combs_cnt; il_pos += VECT_SIZE)
  {
    const u32x pw_r_len = pwlenx_create_combt (combs_buf, il_pos);

    const u32x pw_len = pw_l_len + pw_r_len;

    u32x wordr0[4] = { 0 };
    u32x wordr1[4] = { 0 };
    u32x wordr2[4] = { 0 };
    u32x wordr3[4] = { 0 };

    wordr0[0] = ix_create_combt (combs_buf, il_pos, 0);
    wordr0[1] = ix_create_combt (combs_buf, il_pos, 1);
    wordr0[2] = ix_create_combt (combs_buf, il_pos, 2);
    wordr0[3] = ix_create_combt (combs_buf, il_pos, 3);
    wordr1[0] = ix_create_combt (combs_buf, il_pos, 4);
    wordr1[1] = ix_create_combt (combs_buf, il_pos, 5);
    wordr1[2] = ix_create_combt (combs_buf, il_pos, 6);
    wordr1[3] = ix_create_combt (combs_buf, il_pos, 7);

    if (combs_mode == COMBINATOR_MODE_BASE_LEFT)
    {
      switch_buffer_by_offset_le (wordr0, wordr1, wordr2, wordr3, pw_l_len);
    }

    u32x w_t[16];

    w_t[ 0] = wordl0[0] | wordr0[0];
    w_t[ 1] = wordl0[1] | wordr0[1];
    w_t[ 2] = wordl0[2] | wordr0[2];
    w_t[ 3] = wordl0[3] | wordr0[3];
    w_t[ 4] = wordl1[0] | wordr1[0];
    w_t[ 5] = wordl1[1] | wordr1[1];
    w_t[ 6] = wordl1[2] | wordr1[2];
    w_t[ 7] = wordl1[3] | wordr1[3];
    w_t[ 8] = wordl2[0] | wordr2[0];
    w_t[ 9] = wordl2[1] | wordr2[1];
    w_t[10] = wordl2[2] | wordr2[2];
    w_t[11] = wordl2[3] | wordr2[3];
    w_t[12] = wordl3[0] | wordr3[0];
    w_t[13] = wordl3[1] | wordr3[1];
    w_t[14] = wordl3[2] | wordr3[2];
    w_t[15] = 0;

    u32x a = MYSQL323_A;
    u32x b = MYSQL323_B;
    u32x c = 0;
    u32x d = 0;

    u32 add = 7;

    #define ROUND(v)                              \
    {                                             \
      a ^= (((a & 0x3f) + add) * (v)) + (a << 8); \
      b += (b << 8) ^ a;                          \
      add += v;                                   \
    }

    int i;
    int j;

    for (i = 0, j = 0; i <= (int) pw_len - 4; i += 4, j += 1)
    {
      const u32 wj = w_t[j];

      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
      ROUND ((wj >> 16) & 0xff);
      ROUND ((wj >> 24) & 0xff);
    }

    const u32 wj = w_t[j];

    const u32 left = pw_len - i;

    if (left == 3)
    {
      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
      ROUND ((wj >> 16) & 0xff);
    }
    else if (left == 2)
    {
      ROUND ((wj >>  0) & 0xff);
      ROUND ((wj >>  8) & 0xff);
    }
    else if (left == 1)
    {
      ROUND ((wj >>  0) & 0xff);
    }

    a &= 0x7fffffff;
    b &= 0x7fffffff;

    COMPARE_S_SIMD (a, b, c, d);
  }
}

__kernel void m00200_s08 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
}

__kernel void m00200_s16 (__global pw_t *pws, __global kernel_rule_t *rules_buf, __global comb_t *combs_buf, __global bf_t *bfs_buf, __global void *tmps, __global void *hooks, __global u32 *bitmaps_buf_s1_a, __global u32 *bitmaps_buf_s1_b, __global u32 *bitmaps_buf_s1_c, __global u32 *bitmaps_buf_s1_d, __global u32 *bitmaps_buf_s2_a, __global u32 *bitmaps_buf_s2_b, __global u32 *bitmaps_buf_s2_c, __global u32 *bitmaps_buf_s2_d, __global plain_t *plains_buf, __global digest_t *digests_buf, __global u32 *hashes_shown, __global salt_t *salt_bufs, __global void *esalt_bufs, __global u32 *d_return_buf, __global u32 *d_scryptV_buf, const u32 bitmap_mask, const u32 bitmap_shift1, const u32 bitmap_shift2, const u32 salt_pos, const u32 loop_pos, const u32 loop_cnt, const u32 combs_cnt, const u32 digests_cnt, const u32 digests_offset, const u32 combs_mode, const u32 gid_max)
{
}
